//*** YModem for Delphi v1.00 - (c)1999, EFD Systems ***
//
//This source code is licensed for the private personal use of
//our clients and may not be re-distributed under any circumstances.
//
//THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
//ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
//THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
//PARTICULAR PURPOSE.

{$B-,H+,X+,J-} //Essential directives

unit YModem;

interface

uses Windows, Classes, SysUtils, HyperStr;

type
  TYModemCallBack = procedure(Sender:TObject;Status:AnsiString);

type
  TYModemRX = class(TThread)
  private
    Port:THandle;
    Path:AnsiString;
    StatusCallBack:TYModemCallBack;
    Msg:AnsiString;
  protected
    procedure Execute; override;
    procedure DoCallBack;
  public
    constructor Create(FPort:THandle;FPath:AnsiString;FStatus:TYModemCallBack);
    property Terminated;
    property ReturnValue;
  end;

type
  TYModemTX = class(TThread)
  private
    Port:THandle;
    Files:AnsiString;
    StatusCallBack:TYModemCallBack;
    Msg:AnsiString;
  protected
    procedure Execute; override;
    procedure DoCallBack;
  public
    constructor Create(FPort:THandle;FFiles:AnsiString;FStatus:TYModemCallBack);
    property Terminated;
    property ReturnValue;
  end;

implementation

const
  NUL = #0;
  SOH = #1;                               //Start of header of 128 byte block
  STX = #2;                               //Start of header of 1K block
  EOT = #4;                               //End of transmission
  ACK = #6;                               //Packet OK
  NAK = #21;                              //Packet not OK
  CAN = #24;                              //Cancel
  CRC = #67;                              //CRC transfer
  MaxErrors = 10;                         //max error count before abort
  Tick = 55;                              //short delay
  Tick2 = 110;
  Tick3 = 165;
  PktSize = 1024;
  TimeOut = 3000;

constructor TYmodemRX.Create(FPort:THandle;FPath:AnsiString;FStatus:TYModemCallBack);
begin
  Port := FPort;
  Path := FPath;
  StatusCallBack:=FStatus;
  inherited Create(False);
end;


procedure TYModemRX.DoCallBack;
begin
  StatusCallBack(Self,Msg)
end;


procedure TYModemRX.Execute;
 {Result Codes
 -------------------------------------------------------------------
  >0   - Transfer rate (characters per second), success!!
   0   - Unknown/Out of sync
  -1   - System file error
  -2   - Timeout/Invalid response
  -3   - Too many errors
  -4   - Receiver cancelled
  -5   - Sender cancelled}

var
  FileName,PktHdr,Packet,PktChk,Reply:AnsiString;
  TimeStart,FileLen,BytesRecv,TotalBytesRecv,Errors,LFStatus,PktNum:Integer;
  AbortFlg,CRCFlg,DoneFlg:Boolean;
  C:Char;
  ChkPtr:^Word;
  hF:File;

  procedure ShowStatus(S:AnsiString);
  //--- Relay current status back to main thread
  begin
    Msg:=S;
    if IsConsole then  //Dedicated StatusCallBack function required to avoid
      DoCallBack       //synchronization problems
    else Synchronize(DoCallBack); //Shared StatusCallBack OK
  end;

  function Response:Char;
  //---  Get single character reply from transmitter
  begin
    if ReadComm(Port,Reply)=1 then Result:=Reply[1] else Result:=NUL;
  end;

  procedure RequestIt;
  //--- Request file transfer
  begin
    if PktNum=0 then if CRCFlg then WriteComm(Port,CRC) else WriteComm(Port,NAK);
  end;

  function MakeIt:Boolean;
  //--- Make a file for incoming packets
  var
    I,J:Integer;
  begin
    Result:=False;
    FileName := PChar(Packet);
    if Length(FileName)>0 then begin
      I:=Length(FileName)+2;
      J:=I;
      while IsNumChar(Packet[J]) do Inc(J);
      FileLen := StrToIntDef(CStr(Packet,I,J-I),MAXLONG);
      I := ScanB(FileName,'\',0);
      if I>0 then FileName := RStr(FileName,Length(FileName)-I);
      ShowStatus('Receiving '+FileName+' '+IntToStr(FileLen)+' bytes');
      if Length(Path)>0 then FileName:=Path+FileName;
      try
        AssignFile(hF, FileName);
        Rewrite(hF,1);
        if TimeStart=0 then TimeStart:=GetTickCount;
        LFStatus := 0;
        Result:=True;
        RequestIt;
      except
        AbortFlg:=True;
        LFStatus:=-1;
      end;
    end else begin
      Result:=True;
      DoneFlg:=True;
    end;
  end;

  function WriteIt:Boolean;
  //--- Write a packet to file
  var
    PacketSize:Integer;
  begin
    Result:=False;
    if PktNum=0 then
      Result:=MakeIt
    else begin
      PacketSize:=Length(Packet);
      Inc(BytesRecv,PacketSize);
      if BytesRecv>FileLen then Dec(PacketSize,BytesRecv-FileLen);
      try
        BlockWrite(hF,Packet[1],PacketSize);
        Errors := 0;
        Result:=True;
      except
        AbortFlg:=True;
        LFStatus:=-1;
      end;
    end;
  end;


  procedure NakIt;
  //--- Send Nak
  begin
    ShowStatus('#'+IntToStr(PktNum)+' NG');
    Inc(Errors);
    Sleep(Tick);
    PurgeComm(Port,PURGE_RXCLEAR);      //clear receive buffer
    WriteComm(Port,NAK);
  end;

  procedure AckIt;
  //--- Send Ack
  begin
    ShowStatus('#'+IntToStr(PktNum)+' OK');
    Sleep(Tick);
    WriteComm(Port,ACK);
    Errors:=0;
  end;

  procedure TrashIt;
  //--- Abort transfer and clean up
  begin
    AbortFlg:=True;
    PurgeComm(Port,PURGE_RXCLEAR);
    WriteComm(Port,DupChr(CAN,4));
    if TFileRec(hF).Mode=fmInOut then CloseFile(hF);
  end;

  function NumOk:Boolean;
  //--- Check packet number
  begin
    Result:=False;
    if ReadComm(Port,PktHdr)=Length(PktHdr) then
      Result :=(Ord(PktHdr[1]) = (PktNum and 255)) and
               (Ord(PktHdr[2]) = (255-(PktNum and 255)));
    if Not Result then ShowStatus('Bad packet number!');
  end;

  function CheckOk:Boolean;
  //--- Check packet CRC
  begin
    Result:=False;
    if ReadComm(Port,PktChk)=Length(PktChk) then begin
      if CRCFlg then begin
        ChkPtr:=@PktChk[1];
        Result := CRCXY(0,Packet)=ChkPtr^;
      end else Result:=PktChk[1]=Char(ChkSumXY(Packet));
      if Not Result then ShowStatus('CRC failed!');
    end;
  end;

  function GetIt:Boolean;
  //---  Get Data Packet
  begin
    Result:=False;
    if NumOk then begin
      if ReadComm(Port,Packet)=Length(Packet) then begin
        if CheckOk then Result:=WriteIt;
        Inc(TotalBytesRecv,Length(Packet)+5);
      end else ShowStatus('Read Error');
      if Result then begin
         AckIt;
         Inc(PktNum);
      end else NakIt;
    end else ShowStatus('No packet header');
  end;

  procedure CloseIt;
  //--- Close file and end transfer
  begin
    if PktNum>0 then begin
      PktNum := -1;
      WriteComm(Port,NAK);
      BytesRecv := 0;
      if TFileRec(hF).Mode=fmInOut then CloseFile(hF);
    end else if PktNum<>0 then begin
      PktNum := 0;
      WriteComm(Port,ACK);
    end else Inc(Errors);
  end;

  function TimeIt:Integer;
  //--- Calculate average transfer rate
  begin
    TimeStart := (Integer(GetTickCount)-TimeStart) div 1000;
    Result := TotalBytesRecv div TimeStart;
  end;

begin
  Errors := 0;
  PktNum := 0;
  LFStatus := -2;
  TimeStart := 0;
  FileLen := MAXLONG;
  BytesRecv := 0;
  TotalBytesRecv := 0;
  CRCFlg := True;
  DoneFlg:=False;
  AbortFlg := False;
  Setlength(PktHdr,2);
  Setlength(Packet,128);
  Setlength(PktChk,2);
  Setlength(Reply,1);
  if Length(Path)>0 then begin
    if Path[1]='@' then
      SetLength(Path,0)
    else if Path[Length(Path)]<>'\' then Path:=Path+'\';
  end;
  SetRxTime(Port,TimeOut,30);     //set response wait time
  SetupComm(Port,PktSize+(PktSize SHR 1),128);       //ask for these min. buffer sizes
  PurgeComm(Port,PURGE_RXCLEAR);  //clear receive buffer
  RequestIt;  //send initial transfer request

  //---  Main Reception Loop
  repeat
    C:=Response;
    case C of

      NUL: begin  //timeout
        Inc(Errors);
        if PktNum=0 then begin
          if (Errors=5) then begin
            ShowStatus('Switching to checksum');
            CRCFlg:=False;
            Setlength(PktChk,1);
          end;
          RequestIt;
        end else begin
          ShowStatus('TimeOut');
        end;
        Sleep(Tick2);
      end;

      SOH: begin   //small packet
        SetLength(Packet,128);
        GetIt;
      end;

      STX: begin   //large packet
        SetLength(Packet,PktSize);
        GetIt;
      end;

      CAN: begin   //cancel
        ShowStatus('Sender cancelled');
        LFStatus:=-5;
        TrashIt;
      end;

      EOT: begin   //end of file
        CloseIt;
        RequestIt;
      end;

      else begin    //say what ?
        Inc(Errors);
        if PktNum>0 then begin
          ShowStatus('Unknown response #'+IntToStr(Ord(C)));
          if BytesRecv>=FileLen then CloseIt;
          Sleep(Tick);
        end;
      end;
    end;
  until DoneFlg OR AbortFlg OR Terminated OR (Errors>MaxErrors);

  if DoneFlg then  //calc average transmission rate
    LFStatus := TimeIt
  else begin         //set error FStatus and abort
    if LFStatus=0 then if AbortFlg then LFStatus:=-4 else LFStatus:=-3;
    TrashIt;
  end;

  ReturnValue := LFStatus;

  inherited Terminate;

end;


constructor TYmodemTX.Create(FPort:THandle;FFiles:AnsiString;FStatus:TYModemCallBack);
begin
  Port := FPort;
  Files := FFiles;
  StatusCallBack:=FStatus;
  inherited Create(False);
end;


procedure TYModemTX.DoCallback;
begin
  StatusCallBack(Self,Msg);
end;


procedure TYModemTX.Execute;
 {Result Codes
 -------------------------------------------------------------------
  >0   - Transfer rate (cps); success!
   0   - Unknown/Out of sync
  -1   - System file error
  -2   - Timeout/invalid response
  -3   - Too many errors
  -4   - Receiver cancelled
  -5   - Sender cancelled}

var
  FileName,PktHdr,Packet,PktChk,Reply:AnsiString;
  TimeStart,FileLen,Ptr,BytesSent,TotalBytesSent,Errors,LFStatus,PktNum:Integer;
  AbortFlg,CRCFlg,DoneFlg,Finished:Boolean;
  C:Char;
  ChkPtr:^Word;
  hF:File;

  procedure ShowStatus(S:AnsiString);
  begin
    Msg:=S;
    DoCallBack;
  end;

  function Response:Char;
  //---  Get single character reply from receiver
  begin
    if ReadComm(Port,Reply)=1 then Result:=Reply[1] else Result:=NUL;
  end;

  function OpenIt:Boolean;
  //--- Open file for transmission
  var
    I:Integer;
  begin
    PktNum := 0;
    Setlength(Packet,128);
    FillStr(Packet,1,#0);
    Result:=False;
    if (Ptr>0) and (Ptr<=Length(Files)) then begin
      Finished:=False;
      FileName := ParseWord(Files,' ,;',Ptr);
      try
        AssignFile(hF, FileName);
        Reset(hF,1);
        FileLen := FileSize(hF);
        I := ScanB(FileName,'\',0);
        ShowStatus('Sending '+FileName+' '+IntToStr(FileLen)+' bytes');
        MoveStr(FileName,I+1,Packet,1,Length(FileName));
        MoveStr(IntToStr(FileLen),1,Packet,Length(FileName)-I+2,12);
        LFStatus := 0;
        Result := True;
      except
        AbortFlg:=True;
        LFStatus:=-1;
      end;
    end else DoneFlg:=True;
  end;

  procedure CloseIt;
  //--- Close file and end transfer
  begin
    WriteComm(Port,EOT);
    if Finished then
      Inc(Errors)
    else begin
      Finished:=True;
      BytesSent := 0;
      if TFileRec(hF).Mode=fmInOut then CloseFile(hF);
    end;
  end;

  function ReadIt:Boolean;
  //--- Read a packet from file
  var
    PacketSize:Integer;
  begin
    Result:=False;
    PacketSize:=iMin(PktSize,FileLen-BytesSent);
    if PacketSize<=0 then
      CloseIt
    else begin
      SetLength(Packet,PktSize);
      if PacketSize<Length(Packet) then FillStr(Packet,PacketSize,#26);
      try
        BlockRead(hF,Packet[1],PacketSize);
        Inc(BytesSent,Length(Packet));
        Errors := 0;
        Result:=True;
      except
        AbortFlg := True;
        LFStatus := -1;
      end;
      Inc(PktNum);
    end;
  end;

  procedure TrashIt;
  //--- Abort transfer and clean up
  begin
    PurgeComm(Port,PURGE_RXCLEAR);
    WriteComm(Port,DupChr(CAN,4));
    if TFileRec(hF).Mode=fmInOut then CloseFile(hF);
    AbortFlg:=True;
  end;

  function NumIt:Boolean;
  //--- Fill in packet header
  begin
    Result:=False;
    if PktNum>=0 then begin
      if Length(Packet)=128 then PktHdr[1]:=SOH else PktHdr[1]:=STX;
      PktHdr[2] := Char(PktNum and 255);
      Pkthdr[3] := Char(255-(PktNum and 255));
      Result := WriteComm(Port,PktHdr)=Length(PktHdr);
    end;
  end;

  function CheckIt:Boolean;
  //--- Calc packet CRC
  begin
    if CRCFlg then begin
      ChkPtr:=@PktChk[1];
      ChkPtr^ := CRCXY(0,Packet);
    end else PktChk[1]:=Char(ChkSumXY(Packet));
    Result:=WriteComm(Port,PktChk)=Length(PktChk);
  end;

  function PutIt:Boolean;
  //---  Send Data Packet
  begin
    Result:=False;
    PurgeComm(Port,PURGE_RXCLEAR);  //clear receive buffer
    if NumIt then begin
      if WriteComm(Port,Packet)=Length(Packet) then begin
        Result:=CheckIt;
        Inc(TotalBytesSent,Length(Packet)+5);
      end else ShowStatus('Write failure');
    end;
    Sleep(Tick);  //allow some time for reception overhead
  end;

  function TimeIt:Integer;
  //--- Calculate transfer rate
  begin
    TimeStart := (Integer(GetTickCount)-TimeStart) div 1000;
    Result := TotalBytesSent div TimeStart;
  end;

begin
  Ptr := 1;
  Errors := 0;
  PktNum := -1;
  LFStatus := -2;
  TimeStart := 0;
  FileLen := 0;
  BytesSent := 0;
  TotalBytesSent := 0;
  CRCFlg := True;
  DoneFlg := False;
  Finished := False;
  AbortFlg := False;
  Setlength(PktHdr,3);
  Setlength(Packet,128);
  Setlength(PktChk,2);
  Setlength(Reply,1);
  SetRxTime(Port,TimeOut,30);      //set response wait time
  SetupComm(Port,128,PktSize+(PktSize SHR 1)); //requested min. buffer sizes
  PurgeComm(Port,PURGE_RXCLEAR);  //clear receive buffer
  WriteComm(Port,#13);            //wake 'em up, let 'em know we're coming

  //---  Main Reception Loop
  repeat
    C := Response;
    case C of

      NUL: begin  //timeout
        Inc(Errors);
        ShowStatus('TimeOut');
        if (PktNum>0) and (Errors=MaxErrors) then begin
          if ReadIt then PutIt;  //assume we missed an ACK; last ditch effort
        end;                     //to save the transfer
        Sleep(Tick);
      end;

      ACK: begin   //packet OK
        if Finished then begin
          if DoneFlg then break else PktNum := -1;
        end else begin
          ShowStatus('#'+IntToStr(PktNum)+' OK');
          if ReadIt then PutIt;
        end;
      end;

      NAK: begin   //packet NG
        if Finished then begin
          CloseIt;
        end else begin
          if PktNum<0 then begin
            ShowStatus('Switching to Checksum');
            if TimeStart=0 then TimeStart:=GetTickCount;
            CRCFlg := False;
            SetLength(PktChk,1);
            OpenIt;
          end else begin
            ShowStatus('#'+IntToStr(PktNum)+' NG');
            Inc(Errors);
          end;
          PutIt;
        end;
      end;

      CAN: begin   //cancel
        ShowStatus('Receiver cancelled');
        LFStatus:=-5;
        TrashIt;
      end;

      CRC: begin   //file request
        if PktNum<0 then begin
          if TimeStart=0 then TimeStart:=GetTickCount;
          OpenIt;
          PutIt;
        end;
      end;

      else begin    //say what ?
        ShowStatus('Unknown response '+IntToStr(Ord(C)));
        if BytesSent>=FileLen then CloseIt;
        Sleep(Tick2);
      end;
    end;
  until AbortFlg OR Terminated OR (Errors>MaxErrors);

  if DoneFlg then    //calc average transmission rate
    LFStatus:=TimeIt
  else begin         //set error FStatus and abort
    if LFStatus=0 then if AbortFlg then LFStatus:=-5 else LFStatus:=-3;
    TrashIt;
  end;

  ReturnValue := LFStatus;

  inherited Terminate;

end;

end.
